package edu.northwestern.cs.aqualab.cattle;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.nio.channels.FileChannel;
import java.nio.channels.FileLock;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.UUID;

import edu.northwestern.cs.aqualab.cattle.exception.CattleInitException;
import edu.northwestern.cs.aqualab.cattle.exception.CattleStateException;
import edu.northwestern.cs.aqualab.cattle.logging.Logging;



/**
 * Contains static methods for interacting with the filesystem. There is a
 * configurable limit on total report disk usage, which is enforced by
 * automatic, oldest-first deletion.
 */
public class Filesystem {
    // Reports will be in files named REPORT_PREFIX + hash(body).
    private static final String REPORT_PREFIX = "report-";
    private static final String UUID_FILENAME = "uuid";
    // We maintain exclusive access to the reports directory with an OS-provided
    // advisory lock.
    // XXX: This does not work on older OSes, possibly including OS 10.6 (Snow
    // Leopard). cattle will log a message at level error and assume it is safe
    // to proceed on such platforms.
    private static final String LOCK_FILENAME = "lock";

    // Protects against concurrent access from two different processes.
    private static FileLock sExternalDiskLock = null;
    // To support Java <1.7, we need to keep track of both of these.
    private static FileOutputStream sExternalDiskLockStream = null;
    private static FileChannel sExternalDiskLockChannel = null;

    // We have a single, global lock on disk access. While coarse, I think it'll
    // be fine - we're dealing with small chunks of data, low throughput, and
    // no latency demands.
    private static Object sInternalDiskLock = new Object();
    private static File sRootDir = null;
    private static long sAllowableReportBytes = 10 * 1024 * 1024; // 10 MiB


    /**
     * Generate a filename for a {@link Report}. This filename will always be
     * the same for the same object, but may change between cattle versions.
     */
    static String mkFilename(Report report) {
        return REPORT_PREFIX + report.mUUID.toString();
    }


    /**
     * @return Whether the configured root dir is set and usable.
     */
    public static boolean isRootUsable() {
        return sRootDir != null && sRootDir.isDirectory() && sRootDir.canRead()
               && sRootDir.canWrite();
    }


    /**
     * Set the root directory for all future operations.
     */
    public static void setRoot(String root) {
        sRootDir = new File(root);
        sRootDir.mkdirs();
        if (!isRootUsable()) throw new CattleInitException("Insufficient permissions on " + root);
        if (!lockRoot()) throw new CattleInitException("Couldn't lock " + root);
    }


    /**
     * Acquire an OS-implemented advisory lock on a file relative to the root
     * dir.
     */
    protected static boolean lockRoot() {
        if (sExternalDiskLock != null && sExternalDiskLock.isValid()) {
            Logging.debug("We already have a lock for %s", sRootDir);
            return true;
        }

        File lockFile = new File(sRootDir, LOCK_FILENAME);
        lockFile.deleteOnExit();
        try {
            sExternalDiskLockStream = new FileOutputStream(lockFile);
            sExternalDiskLockChannel = sExternalDiskLockStream.getChannel();

            // tryLock() returns null if and only if it fails to acquire a lock
            // because another process has it. It throws an IOException if we're
            // on an old OS that doesn't support these locks, or if cattle has
            // seriously screwed up. In either case, we have to take our
            // chances.
            sExternalDiskLock = sExternalDiskLockChannel.tryLock();
            if (sExternalDiskLock == null) {
                unlockRoot();
                return false;
            }

            // Explicitly release the lock on shutdown.
            Runtime.getRuntime().addShutdownHook(new Thread() {
                @Override
                public void run() {
                    unlockRoot();
                }
            });
        } catch (IOException ioe) {
            ioe.printStackTrace();
            Logging.error("IOException: %s. Unsupported platform? Tread carefully.");
        }
        return true;
    }


    /**
     * Release the root dir's lock.
     */
    protected static void unlockRoot() {
        // This releases the OS lock and associated resources. The actual lock
        // file is deleted by the JVM due to a .deleteOnExit() in lockRoot().

        if (sExternalDiskLock != null) {
            try {
                sExternalDiskLock.release();
            } catch (IOException ioe) {}
            sExternalDiskLock = null;
        }

        if (sExternalDiskLockChannel != null) {
            try {
                sExternalDiskLockChannel.close();
            } catch (IOException ioe) {}
            sExternalDiskLockChannel = null;
        }

        if (sExternalDiskLockStream != null) {
            try {
                sExternalDiskLockStream.close();
            } catch (IOException ioe) {}
        }
    }


    /**
     * Read all {@link Report Reports} in the given working directory and return
     * them to the caller.
     *
     * @return A Set of Reports
     */
    public static Set<Report> readReports() {
        synchronized (sInternalDiskLock) {
            assert isRootUsable();

            long totalReportBytes = 0;
            // My kingdom for a list comprehension!
            File[] all_files = sRootDir.listFiles();
            Set<File> report_files = new HashSet<File>();
            for (File file : all_files) {
                if (file.getName().startsWith(REPORT_PREFIX)) {
                    report_files.add(file);
                }
            }

            Set<Report> reports = new HashSet<Report>();
            for (File file : report_files) {
                totalReportBytes += file.length();

                ObjectInputStream ois = null;
                Report report = null;
                try {
                    ois = new ObjectInputStream(new FileInputStream(file));
                    report = (Report) ois.readObject();
                } catch (ClassNotFoundException cnfe) {
                    String msg =
                        "readReports(): "
                            + "caught ClassNotFoundException: Something is really, really wrong.";
                    Logging.error(msg);
                    throw new CattleStateException(msg);
                } catch (FileNotFoundException fnfe) {
                    Logging.warn("readReports(): Report disappeared?! %s", file.getName());
                } catch (IOException ioe) {
                    Logging.warn("readReports(): IOException while reading %s: %s", file.getName(),
                                 ioe.getMessage());
                } catch (ClassCastException cce) {
                    Logging
                        .warn("readReports(): Unable to cast report %s to a report. Discarding.",
                              file.getName());
                } finally {
                    if (ois != null) {
                        try {
                            ois.close();
                        } catch (IOException ioe) {}
                    }
                }

                if (report == null) {
                    file.delete();
                    continue;
                }

                assert file.getName().equals(mkFilename(report));
                reports.add(report);
            }

            Logging.debug("readReports(): disk usage: %,d B", totalReportBytes);
            if (totalReportBytes > sAllowableReportBytes) {
                Logging.warn("readReports(): Maximum disk usage exceeded; cleaning.");
                cleanReports(reports);
            }
            return reports;
        }
    }


    /**
     * Write a single report to the appropriate place on disk.
     */
    public static void writeReport(Report report) {
        assert isRootUsable();

        synchronized (sInternalDiskLock) {
            String report_filename = mkFilename(report);
            File report_file = new File(sRootDir, report_filename);
            ObjectOutputStream oos = null;
            try {
                oos = new ObjectOutputStream(new FileOutputStream(report_file));
                oos.writeObject(report);
            } catch (IOException ioe) {
                Logging.warn("writeReport(): Couldn't write report %s to disk. Discarding.",
                             report_filename);
            } finally {
                if (oos != null) {
                    try {
                        oos.close();
                    } catch (IOException ioe) {}
                }
            }
        }
    }


    /**
     * Delete a report from disk. If the report is not already stored on disk,
     * this method will log a warning, but will not throw an exception.
     */
    public static void deleteReport(Report report) {
        assert isRootUsable();

        synchronized (sInternalDiskLock) {
            String report_filename = mkFilename(report);
            File report_file = new File(sRootDir, report_filename);
            if (!report_file.exists()) {
                Logging.warn("deleteReport(): Tried to delete nonexistent %s.", report_filename);
                return;
            }
            report_file.delete();
        }
    }


    /**
     * Delete all reports from disk.
     */
    public static void deleteAllReports() {
        assert isRootUsable();

        synchronized (sInternalDiskLock) {
            Set<Report> reports = readReports();
            for (Report report : reports) {
                deleteReport(report);
            }
        }
    }


    /**
     * Reduce disk usage by deleting some reports. <br>
     * <br>
     * As implemented, this deletes the oldest 10% of reports. This avoids the
     * need to call it on every report once the limit is reached.
     */
    private static void cleanReports(Set<Report> reportSet) {
        List<Report> reports = new ArrayList<Report>(reportSet.size());
        reports.addAll(reportSet);
        Collections.sort(reports, new Comparator<Report>() {
            public int compare(Report one, Report another) {
                Long oneCreatedAt = one.createdAt;
                Long anotherCreatedAt = another.createdAt;
                return oneCreatedAt.compareTo(anotherCreatedAt);
            }
        });

        int numToDelete = (int) Math.ceil((double) reports.size() / 10);
        for (int i = 0; i < numToDelete; i++) {
            Report report = reports.get(i);
            Logging.warn("cleanReports(): Deleting for space: %s", report);
            deleteReport(report);
        }
    }


    /**
     * Get the persistent, unique client ID from disk.
     *
     * @return The saved client UUID, or null if it could not be read.
     */
    public static UUID readClientUUID() {
        assert isRootUsable();

        synchronized (sInternalDiskLock) {
            File UUIDFile = new File(sRootDir, UUID_FILENAME);
            UUID clientUUID = null;
            ObjectInputStream ois = null;
            try {
                ois = new ObjectInputStream(new FileInputStream(UUIDFile));
                clientUUID = (UUID) ois.readObject();
            } catch (FileNotFoundException fnfe) {
                Logging.debug("readClientUUID(): No saved UUID found.");
            } catch (IOException ioe) {
                Logging.warn("readClientUUID(): IOError: " + ioe.getMessage());
            } catch (ClassNotFoundException cnfe) {
                String msg =
                    "readClientUUID(): ClassNotFoundException: Something is really, really wrong.";
                Logging.error(msg);
            } catch (ClassCastException cce) {
                Logging.warn("readClientUUID(): Unable to cast UUID file to a UUID. Discarding.");
                UUIDFile.delete();
            } finally {
                if (ois != null) {
                    try {
                        ois.close();
                    } catch (IOException ioe) {}
                }
            }
            return clientUUID;
        }
    }


    /**
     * Set the persistent, unique client ID on disk.
     */
    public static void writeClientUUID(UUID clientUUID) {
        assert isRootUsable();

        synchronized (sInternalDiskLock) {
            File UUIDFile = new File(sRootDir, UUID_FILENAME);

            ObjectOutputStream oos = null;
            try {
                oos = new ObjectOutputStream(new FileOutputStream(UUIDFile));
                oos.writeObject(clientUUID);
            } catch (IOException ioe) {
                Logging.error("writeClientUUID(): Couldn't write UUID to disk.");
            } finally {
                if (oos != null) {
                    try {
                        oos.close();
                    } catch (IOException ioe) {}
                }
            }
        }
    }


    /**
     * Set the disk usage limit you want the library to enforce. Note that this
     * is not a strict limit - large reports and other metadata may result in
     * slightly higher usage.
     */
    protected static void setDiskUsageLimit(long bytes) {
        sAllowableReportBytes = bytes;
        Logging.debug("setDiskUsageLimit: %,d B", bytes);
    }
}
